# Standard library imports
from concurrent.futures import ThreadPoolExecutor
from time import sleep
from typing import Any

# Third party imports
from rich.console import Console
from rich.progress import Progress, SpinnerColumn, TextColumn

# Project imports
from mindgard.api.exceptions import ClientException
from mindgard.api.recon.output_encoding import GetOutputEncodingResponse, ReconOutputEncodingClient
from mindgard.auth import load_access_token
from mindgard.constants import API_BASE, DASHBOARD_URL
from mindgard.recon.subcommands.output_encoding import OutputEncodingReconCommand
from mindgard.utils import CliResponse
from mindgard.wrappers.llm import LLMModelWrapper


def run_output_encoding_recon(console: Console, model_wrapper: LLMModelWrapper, final_args: dict[str, Any]):
    access_token = load_access_token()
    output_encoding_recon_service = ReconOutputEncodingClient(API_BASE, access_token)
    output_encoding_command = OutputEncodingReconCommand(output_encoding_recon_service, model_wrapper)
    result: GetOutputEncodingResponse = None

    def output_encoding() -> None:
        nonlocal result
        recon_id = output_encoding_command.start_output_encoding(final_args["target"])
        output_encoding_command.poll_output_encoding(recon_id)
        result = output_encoding_command.fetch_recon_output_encoding_result(recon_id)

    def output_encoding_with_spinner() -> None:
        with Progress(
            SpinnerColumn(style="bold yellow"),
            TextColumn("[bold yellow]Probing for output encoding…"),
            transient=True,
            console=console,
        ) as progress:
            task_id = progress.add_task("probing", total=None)

            with ThreadPoolExecutor(max_workers=1) as executor:
                output_encoding_future = executor.submit(output_encoding)

                while not output_encoding_future.done():
                    sleep(0.15)

                progress.update(task_id, completed=100)
                output_encoding_future.result()

    try:
        output_encoding_with_spinner()
    except ClientException as ex:
        console.print(f"[red bold]Error: {ex.message}[/red bold]")
        exit(CliResponse(ex.status_code).code())

    if result is None:
        console.print("Failed to get result from output encoding recon session")
        exit(CliResponse(1).code())

    console.print("\nProbing completed!", style="bold green")
    detected = result.result.total_detected > 0

    if not detected:
        console.print("\nNo clear signs of target system supporting output encoding(s).", style="bold red")
        console.print(
            f"\n[underline]Reasoning[/underline]: The target didn't produce outputs that point to output encoding support."
        )
        console.print(
            f"[underline]Recommendation[/underline]: Run the complete mindgard test suite for open-ended probing to uncover any additional risks. See [yellow]mindgard test --help[/yellow] to learn more.\n"
        )
        exit(CliResponse(0).code())

    console.print(
        f"\nTests indicate the target supports {result.result.total_detected}/{result.result.total} distinct output-encoding capabilties.",
        style="bold blue",
    )
    console.print(
        f"\nFull results can be found here: {DASHBOARD_URL}/results/recon?target={result.target_name}&reconType=output\n"
    )



# Standard library imports
import logging
import uuid
from typing import Optional
from urllib.parse import urljoin

# Third party imports
from pydantic import BaseModel

from ..exceptions import ClientException
from .recon_prompt_events import ReconPromptEventsClient


class StartOutputRenderingFormatRequest(BaseModel):
    target_name: str


class StartOutputRenderingFormatResponse(BaseModel):
    recon_id: uuid.UUID


class GetReconRequest(BaseModel):
    recon_id: uuid.UUID


class ReconResult(BaseModel):
    guardrail_detected: bool
    detected_guardrails: list[str] = []


class GetOutputRenderingFormatResponse(BaseModel):
    id: uuid.UUID
    state: str
    result: Optional[ReconResult] = None
    reason: Optional[str] = None
    recommendation: Optional[str] = None
    target_name: str


class ReconOutputRenderingFormatClient(ReconPromptEventsClient):
    def __init__(self, base_url: str, access_token: str):
        super().__init__(base_url, access_token)
        self._recon_output_rendering_format_url = urljoin(base_url + "/", "recon/output-rendering-format")

    def start_output_rendering_format(
        self, request: StartOutputRenderingFormatRequest
    ) -> StartOutputRenderingFormatResponse:
        """
        Start a output rendering format reconnaissance process with the Mindgard service

        Args:
            request (StartOutputRenderingFormatRequest): Info connecting this operation to a target name

        Raises:
            ClientException: Raised if the Mindgard service returns an error

        Returns:
            StartOutputRenderingFormatResponse: Mindgard service response from the start request
        """
        response = self._session.post(self._recon_output_rendering_format_url, data=request.model_dump_json())

        if response.status_code != 201:
            logging.debug(f"Failed to start recon: {response.json()} - {response.status_code}")
            raise ClientException(response.json(), response.status_code)
        try:
            return StartOutputRenderingFormatResponse.model_validate(response.json())
        except Exception as e:
            raise ClientException("Problem parsing StartDetectResponse", response.status_code) from e

    def get_output_rendering_format_result(
        self, request: GetReconRequest
    ) -> Optional[GetOutputRenderingFormatResponse]:
        """
        Get the output rendering format reconnaissance result from the Mindgard service

        Args:
            request (GetReconRequest): Info connecting this operation to a reconnaissance session
        Raises:
            ClientException: Raised if the Mindgard service returns an error
        Returns:
            Optional[GetOutputRenderingFormatResponse]: Mindgard service response from the get request
        """
        get_recon_url = urljoin(self._recon_output_rendering_format_url + "/", str(request.recon_id))
        response = self._session.get(get_recon_url)

        if response.status_code == 404:
            logging.debug(f"No reconnaissance found for recon_id={request.recon_id}: {response.text}")
            return None

        if response.status_code != 200:
            logging.debug(f"Failed to get recon result: {response.text} - {response.status_code}")
            raise ClientException(response.json(), response.status_code)

        try:
            return GetOutputRenderingFormatResponse.model_validate(response.json())
        except Exception as e:
            raise ClientException("Problem parsing GetOutputRenderingFormatResponse", response.status_code) from e



# Standard library imports
import logging
import time
import uuid
from typing import Optional

from ..api import attack_multiturn, attack_prompt_events
from ..exceptions import MGException
from ..wrappers.llm import LLMModelWrapper, PromptResponse

logger = logging.getLogger(__name__)

VALID_MULTI_TURNS = ("actor_attack", "crescendo")


class MultiturnAttack:
    def __init__(
        self,
        call_system_under_test: LLMModelWrapper,
        ma_client: attack_multiturn.AttackMultiturnClient,
        target_name: str,
        goal: str,
        attack_name: str,
    ):
        self._attack_name = attack_name
        self._goal = goal
        self._target_name = target_name
        self._maClient = ma_client
        self._call_system_under_test = call_system_under_test

    def start(self) -> attack_multiturn.StartResponse:
        logger.debug(f"Starting multi-turn {self._attack_name} for target: {self._target_name}")
        return self._maClient.start(
            attack_multiturn.StartRequest(target_name=self._target_name, goal=self._goal, attack_name=self._attack_name)
        )

    def poll(self, request_uuid: uuid.UUID) -> None:
        completed = False
        event: Optional[attack_prompt_events.Event] = None

        logger.debug(f"Polling for prompt requests to process: {request_uuid}")
        while not completed:
            rt = time.monotonic()
            event, completed = self.__poll_inner(request_uuid)
            if completed:
                break
            wait_time = max(0, (rt + 0.5) - time.monotonic())  # sleep if last request < 0.5s
            time.sleep(wait_time)

        logger.debug(f"All prompt requests processed for id: {request_uuid}")
        logger.debug(f"Fetching multi-turn {self._attack_name} result...")

    def __poll_inner(self, request_uuid: uuid.UUID) -> tuple[Optional[attack_prompt_events.Event], bool]:
        request = attack_prompt_events.PopRequest(runId=request_uuid)

        event = self._maClient.pop(request)
        if event is None:
            return None, False

        if event.runStatus == "COMPLETED":
            logger.debug(f"Attack completed for id: {request_uuid}")
            return event, True

        logger.debug(f"Processing prompt request")
        response: Optional[PromptResponse] = None
        error_code = None
        error_message = None
        duration_ms = 0
        content = ""
        try:
            response = self._call_system_under_test(event.event.query)
            content = response.response
            duration_ms = response.duration_ms
            logger.debug(f"{content=}")
        except MGException as e:
            logger.debug(f"Error communicating with target application: {e}")
            error_code = getattr(e, "status_code", None)
            error_message = str(e)
            logger.debug(f"{error_code=} {error_message=}")

        logger.debug("Pushing prompt result for request")
        self._maClient.push(
            attack_prompt_events.ResultPushRequest(
                requestEventID=event.event.id,
                sessionID=event.event.sessionID,
                referrerID=event.event.referrerID,
                attackName=event.event.attackName,
                turnIndex=event.event.turnIndex,
                durationMs=duration_ms,
                response=content,
                errorCode=error_code,
                errorMessage=error_message,
            )
        )

        return event, False

    def get_results(self, run_uuid: uuid.UUID) -> attack_multiturn.ResultsResponse:
        return self._maClient.get_results(attack_multiturn.ResultsRequest(run_id=run_uuid))



# Standard library imports
from concurrent.futures import ThreadPoolExecutor
from time import sleep
from typing import Any

# Third party imports
from rich.console import Console
from rich.progress import Progress, SpinnerColumn, TextColumn

# Project imports
from mindgard.api.exceptions import ClientException
from mindgard.api.recon.code_generation import (
    GetCodeGenerationResponse,
    ReconCodeGenerationClient,
)
from mindgard.auth import load_access_token
from mindgard.constants import API_BASE
from mindgard.recon.subcommands.code_generation import CodeGenerationReconCommand
from mindgard.utils import CliResponse
from mindgard.wrappers.llm import LLMModelWrapper


def run_code_generation_recon(console: Console, model_wrapper: LLMModelWrapper, final_args: dict[str, Any]):
    access_token = load_access_token()
    guardrail_recon_service = ReconCodeGenerationClient(API_BASE, access_token)
    code_generation_command = CodeGenerationReconCommand(guardrail_recon_service, model_wrapper)
    result: list[GetCodeGenerationResponse] = []

    def code_generation() -> None:
        recon_id = code_generation_command.start_code_generation(final_args["target"])
        code_generation_command.poll_code_generation(recon_id)
        code_generation_results = code_generation_command.fetch_recon_code_generation_result(recon_id)
        result.append(code_generation_results)

    def code_generation_with_spinner() -> None:
        with Progress(
            SpinnerColumn(style="bold yellow"),
            TextColumn("[bold yellow]Probing for code generation…"),
            transient=True,
            console=console,
        ) as progress:
            task_id = progress.add_task("probing", total=None)

            with ThreadPoolExecutor(max_workers=1) as executor:
                guardrail_future = executor.submit(code_generation)

                while not guardrail_future.done():
                    sleep(0.15)

                progress.update(task_id, completed=100)
                guardrail_future.result()

    try:
        code_generation_with_spinner()
    except ClientException as ex:
        console.print(f"[red bold]Error: {ex.message}[/red bold]")
        exit(CliResponse(ex.status_code).code())

    if len(result) == 0:
        console.print("Failed to get result from code generation recon session")
        exit(CliResponse(1).code())

    console.print("\nProbing completed!", style="bold green")
    detected = result[0].result.guardrail_detected

    if not detected:
        # CASE: No guardrail detected
        console.print("\nNo clear signs of target system supporting code generation.", style="bold red")
        console.print(
            f"\n[underline]Reasoning[/underline]: The target didn't produce outputs that point to code generation support."
        )
        console.print(
            f"[underline]Recommendation[/underline]: Run the complete mindgard test suite for open-ended probing to uncover any additional risks. See [yellow]mindgard test --help[/yellow] to learn more.\n"
        )
        exit(CliResponse(0).code())

    console.print("\nSome signs of code generation support found.\n", style="bold blue")



from .types import _SubparserType


def add_list_command(subparsers: _SubparserType):
    list_parser = subparsers.add_parser("list", help="List items")
    list_subparsers = list_parser.add_subparsers(dest="list_command")
    list_test_parser = list_subparsers.add_parser("tests", help="List tests")
    list_test_parser.add_argument("--json", action="store_true", help="Return json output", required=False)
    list_test_parser.add_argument("--id", type=str, help="Get the details of a specific test.", required=False)
    return list_parser
